/*
 * Copyright (C) 2006 Changbeom Ahn
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <iostream>
#include <string>
#include <sstream>
#include <vector>
#include <iterator>
#include <fstream>
#include <map>
#include <set>


namespace {


	std::string const PREFIX	= "ti_hangul_";
	std::string const NO_GROUP	= "nogroup";


}


typedef unsigned short uc16;
typedef std::vector<uc16> Uc16Vec;

struct Group {
	Group(std::string const&, Uc16Vec const&);
	std::string name;
	Uc16Vec members;
	size_t minIndex, maxIndex;
};

typedef std::vector<Group> Groups;
typedef std::map<std::string, size_t> GroupMap;

struct State {
	size_t next1;
	size_t next2;
	size_t next3;
	size_t next4;
};

typedef std::vector<State> States;
typedef std::map<Uc16Vec, std::map<std::string, size_t> > StateMap;

typedef std::vector<std::string> StrVec;

typedef std::map<std::string, std::string> AliasMap;

typedef std::map<Uc16Vec, std::string> Utf8Map;

typedef std::map<size_t, std::string> TextMap;

struct Export {
	Export(std::string const& s, std::string const& n) :
		state(s), name(n) {}
	std::string state, name;
};

typedef std::vector<Export> Exports;


Groups groups;
GroupMap groupMap;

States states;
StateMap stateMap;

AliasMap aliasMap;

Utf8Map utf8Map;

TextMap textMap;

Exports exports;


std::string
toupper(std::string const& s)
{
	std::string result;
	for (size_t i = 0; i < s.size(); ++i) {
		result.push_back(toupper(s[i]));
	}
	return result;
}


std::string
u2s(Uc16Vec const& u)
{
	Utf8Map::iterator it = utf8Map.find(u);
	if (it != utf8Map.end()) {
		return (*it).second;
	}

	std::ostringstream oss;
	oss << std::hex;

	for (Uc16Vec::const_iterator it = u.begin(); it != u.end(); ++it) {
		oss << "\\x" << *it;
	}

	return oss.str();
}


Group::Group(std::string const& nm, Uc16Vec const& mem) :
	name(nm), members(mem), minIndex(states.size() - members.size()),
	maxIndex(states.size() - 1)
{
}


void
removeComment(std::string& line)
{
	std::string::size_type i = line.find('#');
	if (i == std::string::npos) {
		return;
	}
	line.erase(i);
}


uc16 append(uc16 u, std::string::const_iterator& it)
{
	return (u << 6) | (*++it & 0x3F);
}


Uc16Vec
decodeUtf8(std::string const& utf8)
{
	Uc16Vec result;

	typedef std::string::const_iterator Iter;
	for (Iter it = utf8.begin(); it != utf8.end(); ++it) {
		if ((*it & 0x80) == 0) {
			result.push_back(*it);
		} else if ((*it & 0xE0) == 0xC0) {
			result.push_back(append(*it & 0x1F, it));
		} else if ((*it & 0xF0) == 0xE0) {
			result.push_back(append(append(*it & 0x0F, it), it));
		} else {
			std::cerr << "Unknown UTF-8 sequence: " <<
				utf8 << std::endl;
			exit(EXIT_FAILURE);
		}
	}

	utf8Map[result] = utf8;

	return result;
}


size_t
addState(Uc16Vec const& name, std::string const& group = NO_GROUP)
{
	size_t index = states.size();
	stateMap[name][group] = index;
	std::clog << "addState: index=" << index << " group=" << group <<
		" name=" << u2s(name) << std::endl;

	states.push_back(State());

	return index;
}


size_t
getStateIndex(std::string name)
{
	std::string group = NO_GROUP;
	std::string::size_type i = name.find('.');
	if (i != std::string::npos) {
		group = name.substr(0, i);
		name = name.substr(i + 1);
		//std::clog << "getStateIndex: group=" << group <<
		//	" name=" << name << std::endl;

		AliasMap::iterator it = aliasMap.find(group);
		if (it != aliasMap.end()) {
			group = (*it).second;
			//std::clog << "getStateIndex: group=" <<
			//	group << std::endl;
		}
	}

	Uc16Vec ucName = decodeUtf8(name);
	//std::clog << "getStateIndex: ucName=" << u2s(ucName) << std::endl;
	StateMap::iterator itState = stateMap.find(ucName);
	if (itState == stateMap.end()) {
		std::clog << "getStateIndex: adding state: " <<
			name << std::endl;
		return addState(ucName, group);
	}

	typedef std::map<std::string, size_t> Map;
	Map& m = (*itState).second;
	Map::iterator it;

	switch (m.size()) {
		case 1:
			it = m.begin();
			break;
		default:
			it = m.find(group);
			if (it != m.end()) {
				break;
			}

			std::cerr << "getStateIndex: '" << name <<
				"' is ambiguous" << std::endl <<
				"candidates:";

			for (Map::iterator i = m.begin(); i != m.end(); ++i) {
				std::cerr << ' ' << (*i).first << '.' << name;
			}

			std::cerr << std::endl;

			exit(EXIT_FAILURE);
	}

	return (*(it)).second;
}


State&
getState(std::string const& name)
{
	return states[getStateIndex(name)];
}


void
addAlias(std::string const& alias, std::string const& name)
{
	aliasMap[alias] = name;
}


void
writeGroup(Group const& group)
{
	std::ostringstream oss;
	oss << PREFIX << group.name << "_tab";
	std::string tabName = oss.str();

	std::cout << "uc16 const " << tabName << "[] = {";
	Uc16Vec const& m = group.members;
	std::cout << std::hex;
	for (Uc16Vec::const_iterator j = m.begin(); j != m.end(); ++j) {
		std::cout << "0x" << *j << ", ";
	}
	std::cout << "};" << std::endl;

	std::cout << "#define " <<
		toupper(PREFIX) << "NUM_" << toupper(group.name) << "S "
		"(sizeof(" << tabName << ") / "
		"sizeof(*" << tabName << "))" << std::endl;

	std::cout << std::dec << "#define " <<
		toupper(PREFIX) << "IS_" << toupper(group.name) << "(x) "
		"(" << group.minIndex << " <= (x) && "
		"(x) <= " << group.maxIndex << ")" << std::endl;

	std::cout << "#define " <<
		toupper(PREFIX) << toupper(group.name) << "_INDEX(x) "
		"((x) - " << group.minIndex << ")" << std::endl;
}


void
addGroup(std::string const& name, std::string const& members)
{
	std::clog << "group: name=" << name <<
		" members=" << members << std::endl;

	Uc16Vec memVec = decodeUtf8(members);
	for (Uc16Vec::iterator it = memVec.begin(); it != memVec.end(); ++it) {
		Uc16Vec v;
		v.push_back(*it);
		addState(v, name);
	}

	if (groupMap.find(name) != groupMap.end()) {
		std::cerr << "Group name already exists: " << name << std::endl;
		exit(EXIT_FAILURE);
	}
	groupMap[name] = groups.size();

	groups.push_back(Group(name, memVec));
}


void
evalGroup(std::istream& in)
{
	std::string name, members;
	in >> name >> members;

	addGroup(name, members.substr(1, members.size() - 2));
}


StrVec
split(std::string const& s, char sep)
{
	StrVec result;

	std::string::size_type i = 0;
	while (true) {
		std::string::size_type j = s.find(sep, i);
		if (j == std::string::npos) {
			break;
		}

		result.push_back(s.substr(i, j - i));
		i = j + 1;
	}

	if (i < s.size()) {
		result.push_back(s.substr(i));
	}

	return result;
}


void
updateState(std::string const& name,
		std::string const& next1, std::string const& next2,
		std::string const& next3, std::string const& next4)
{
	//std::clog << "updateState: next1" << std::endl;
	getState(name).next1 = getStateIndex(next1);
	//std::clog << "updateState: next2" << std::endl;
	getState(name).next2 = getStateIndex(next2);
	//std::clog << "updateState: next3" << std::endl;
	getState(name).next3 = getStateIndex(next3);
	//std::clog << "updateState: next4" << std::endl;
	getState(name).next4 = getStateIndex(next4);
}


void
evalTransition(std::istream& in)
{
	std::string names, next1, next2, next3, next4;
	char skip;
	in >> names >> skip >> next1 >> next2 >> next3 >> next4;
	//std::cout << "evalTransition: names=" << names <<
	//	", next1=" << next1 << ", next2=" << next2 <<
	//	", next3=" << next3 << ", next4=" << next4 <<
	//	std::endl;

	StrVec nameVec = split(names, '/');
	//if (nameVec.size() > 0) {
	//	copy(nameVec.begin(), nameVec.end(),
	//			std::ostream_iterator<std::string>(
	//				std::cout, " "));
	//	std::cout << std::endl;
	//}

	for (StrVec::iterator it = nameVec.begin(); it != nameVec.end(); ++it) {
		updateState(*it, next1, next2, next3, next4);
	}
}


void
evalAlias(std::istream& in)
{
	std::string alias, name;

	in >> alias >> name;

	addAlias(alias, name);
}


void
evalText(std::istream& in)
{
	std::string name, value;

	in >> name >> value;

	textMap[getStateIndex(name)] = value.substr(1, value.size() - 2);
}


void
evalExport(std::istream& is)
{
	std::string state, name;
	is >> state >> name;
	exports.push_back(Exports::value_type(state, name));
}


void
eval(std::string& line)
{
	removeComment(line);
	//std::cout << "nc: " << line << std::endl;

	std::istringstream in(line);
	std::string token;
	in >> token;
	if (token.size() == 0) {
		//std::cout << "parse: no token, skipping..." << std::endl;
		return;
	} else if (token == "group") {
		evalGroup(in);
	} else if (token == "alias") {
		evalAlias(in);
	} else if (token == "text") {
		evalText(in);
	} else if (token == "transition") {
		evalTransition(in);
	} else if (token == "export") {
		evalExport(in);
	} else {
		std::cerr << "eval: unknown token: " << token << std::endl;
	}
}


void
writeOutput()
{
	std::cout << "/* Do NOT edit. Generated from tables.in. "
		"You should edit it or gentabs.cpp, not this file. */" <<
		std::endl;
	std::cout << "typedef char " << PREFIX << "index_t;" << std::endl;

	for (Groups::iterator it = groups.begin(); it != groups.end(); ++it) {
		writeGroup(*it);
	}

	std::cout << std::dec;

	std::cout << "struct {\n" <<
		"\t" << PREFIX << "index_t next1;\n" <<
		"\t" << PREFIX << "index_t next2;\n" <<
		"\t" << PREFIX << "index_t next3;\n" <<
		"\t" << PREFIX << "index_t next4;\n" <<
		"} const " << PREFIX << "state_tab[] = {\n";
	for (States::iterator it = states.begin(); it != states.end(); ++it) {
		//std::clog << "writeOutput: " <<
		//	utf8Map[(*it).next1] << ", " <<
		//	utf8Map[(*it).next2] << ", " <<
		//	utf8Map[(*it).next3] << ", " <<
		//	utf8Map[(*it).next4] << std::endl;
		std::cout << "\t{" <<
			(*it).next1 << ", " <<
			(*it).next2 << ", " <<
			(*it).next3 << ", " <<
			(*it).next4 << ", " <<
			"},\n";
		if ((*it).next1 == 0 && (*it).next2 == 0 && (*it).next3 == 0 &&
				(*it).next4 == 0) {
			std::cerr << "writeOutput: warning: "
				"no custom transition defined for state #" <<
				(&(*it) - &states[0]) << std::endl;
		}
	}
	std::cout << "};" << std::endl;

	size_t offset = (*textMap.begin()).first;
	std::string names;
	typedef TextMap::iterator TextMapIter;

	for (TextMapIter it = textMap.begin(); it != textMap.end(); ++it) {
		std::ostringstream oss;
		oss << PREFIX << "text" << std::dec << (*it).first - offset;
		std::string name = oss.str();
		names += name + ", ";

		std::cout << "uc16 const " << name << "[] = {";
		std::cout << std::hex;

		Uc16Vec u = decodeUtf8((*it).second);

		for (Uc16Vec::iterator it = u.begin(); it != u.end(); ++it) {
			std::cout << "0x" << *it << ", ";
		}

		std::cout << "0};" << std::endl;
	}

	std::cout << "uc16 const* " << PREFIX << "text[] = {" << names << "};"
		<< std::endl;

	std::cout << "#define " << toupper(PREFIX) << "TEXT(x) " <<
		PREFIX << "text[(x) - " << offset << "]" << std::endl;

	std::cout << std::dec;
	typedef Exports::iterator ExportsIter;
	for (ExportsIter it = exports.begin(); it != exports.end(); ++it) {
		std::cout << "#define " << toupper(PREFIX) << (*it).name <<
			' ' << getStateIndex((*it).state) << std::endl;
	}
}


void
evalStream(std::istream& is)
{
	while (is) {
		std::string line;
		getline(is, line);
		//std::clog << line << std::endl;
		eval(line);
	}
}


int
main()
{
	std::ifstream ifs("tables.in");
	evalStream(ifs);
	writeOutput();

	/*
	Uc16Vec v1 = decodeUtf8("끝.ㄷ");

	Uc16Vec v3 = decodeUtf8("ㄱㄴㄷ");

	Uc16Vec v4;
	v4.push_back(v3[2]);

	Uc16Vec v2 = decodeUtf8("끝");
	v2.push_back('.');
	copy(v4.begin(), v4.end(), std::back_inserter(v2));

	std::cout << std::hex;

	std::cout << "v1: ";
	copy(v1.begin(), v1.end(), std::ostream_iterator<uc16>(std::cout));
	std::cout << std::endl;

	std::cout << "v2: ";
	copy(v2.begin(), v2.end(), std::ostream_iterator<uc16>(std::cout));
	std::cout << std::endl;
	*/

	return 0;
}
